@@ -44,6 +44,18 @@ struct {
     unsigned int version;
     StringView notes;
 } constexpr version_notes[] = { {
+        20230805,
+        "» Fix FreeBSD/MacOS clang compilation\n"
+    }, {
+        20230729,
+        "» {+b}<a-u>{} and {+b}<a-U>{} now undo/redo selection changes; "
+        "the previous meaning of moving in history tree has been moved to "
+        "{+b}<c-j>{} and {+b}<c-k>{}\n"
+        "» {+u}%exp\\{...}{} expansions provide flexible quoting for expanded "
+        "strings (as double quoted strings)\n"
+        "» {+u}show-matching -previous{} switch\n"
+        "» {+b}<c-g>{} to cancel current operation\n"
+    }, {
         20221031,
         "» {+b}<esc>{} does not end macro recording anymore, use {+b}Q{}\n"
         "» pipe commands do not append final end-of-lines anymore\n"
@@ -136,24 +148,24 @@ void show_startup_info(Client* local_client, int last_version)
 {
     const Face version_face{Color::Default, Color::Default, Attribute::Bold};
     DisplayLineList info;
-    for (auto note : version_notes)
+    for (auto [version, notes] : version_notes)
     {
-        if (note.version and note.version <= last_version)
+        if (version and version <= last_version)
             continue;
 
-        if (not note.version)
+        if (not version)
             info.push_back({"• Development version", version_face});
         else
         {
-            const auto year = note.version / 10000;
-            const auto month = (note.version / 100) % 100;
-            const auto day = note.version % 100;
+            const auto year = version / 10000;
+            const auto month = (version / 100) % 100;
+            const auto day = version % 100;
             info.push_back({format("• Kakoune v{}.{}{}.{}{}",
                                    year, month < 10 ? "0" : "", month, day < 10 ? "0" : "", day),
                             version_face});
         }
 
-        for (auto&& line : note.notes | split<StringView>('\n'))
+        for (auto&& line : notes | split<StringView>('\n'))
             info.push_back(parse_display_line(line, GlobalScope::instance().faces()));
     }
     if (not info.empty())
@@ -349,6 +361,10 @@ static const EnvVarDesc builtin_env_vars[] = { {
                      transform([&](const Selection& s) -> String {
                          return to_string(char_length(context.buffer(), s));
                      }) | gather<Vector<String>>(); }
+    }, {
+        "selection_count", false,
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.selections().size())}; }
     }, {
         "window_width", false,
         [](StringView name, const Context& context) -> Vector<String>
@@ -563,7 +579,8 @@ void register_options()
                        "    terminal_wheel_scroll_amount   int\n"
                        "    terminal_shift_function_key    int\n"
                        "    terminal_padding_char          codepoint\n"
-                       "    terminal_padding_fill          bool\n",
+                       "    terminal_padding_fill          bool\n"
+                       "    terminal_info_max_width        int\n",
                        UserInterface::Options{});
     reg.declare_option("modelinefmt", "format string used to generate the modeline",
                        "%val{bufname} %val{cursor_line}:%val{cursor_char_column} {{context_info}} {{mode_info}} - %val{client}@[%val{session}]"_str);
@@ -621,6 +638,7 @@ std::unique_ptr<UserInterface> make_ui(UIType ui_type)
         void set_cursor(CursorMode, DisplayCoord) override {}
         void refresh(bool) override {}
         void set_on_key(OnKeyCallback) override {}
+        void set_on_paste(OnPasteCallback) override {}
         void set_ui_options(const Options&) override {}
     };
 
@@ -750,7 +768,7 @@ enum class ServerFlags
 constexpr bool with_bit_ops(Meta::Type<ServerFlags>) { return true; }
 
 int run_server(StringView session, StringView server_init,
-               StringView client_init, Optional<BufferCoord> init_coord,
+               StringView client_init, StringView init_buffer, Optional<BufferCoord> init_coord,
                ServerFlags flags, UIType ui_type, DebugFlags debug_flags,
                ConstArrayView<StringView> files)
 {
@@ -832,9 +850,7 @@ int run_server(StringView session, StringView server_init,
 
     if (not files.empty()) try
     {
-        // create buffers in reverse order so that the first given buffer
-        // is the most recently created one.
-        for (auto& file : files | reverse())
+        for (auto& file : files)
         {
             try
             {
@@ -863,7 +879,7 @@ int run_server(StringView session, StringView server_init,
         if (not server.is_daemon())
         {
             local_client = client_manager.create_client(
-                 create_local_ui(ui_type), getpid(), {}, get_env_vars(), client_init, std::move(init_coord),
+                 create_local_ui(ui_type), getpid(), {}, get_env_vars(), client_init, init_buffer, std::move(init_coord),
                  [](int status) { local_client_exit = status; });
 
             if (startup_error and local_client)
@@ -884,12 +900,17 @@ int run_server(StringView session, StringView server_init,
             // Loop so that eventual inputs happening during the processing are handled as
             // well, avoiding unneeded redraws.
             bool allow_blocking = not client_manager.has_pending_inputs();
-            while (event_manager.handle_next_events(EventMode::Normal, nullptr, allow_blocking))
+            try
             {
-                if (client_manager.process_pending_inputs())
-                    break;
-                allow_blocking = false;
+                while (event_manager.handle_next_events(EventMode::Normal, nullptr, allow_blocking))
+                {
+                    if (client_manager.process_pending_inputs())
+                        break;
+                    allow_blocking = false;
+                }
             }
+            catch (const cancel&) {}
+
             client_manager.process_pending_inputs();
 
             client_manager.clear_client_trash();
@@ -1069,29 +1090,28 @@ int main(int argc, char* argv[])
     set_signal_handler(SIGTTOU, SIG_IGN);
 
     const ParameterDesc param_desc{
-        SwitchMap{ { "c", { true,  "connect to given session" } },
-                   { "e", { true,  "execute argument on client initialisation" } },
-                   { "E", { true,  "execute argument on server initialisation" } },
-                   { "n", { false, "do not source kakrc files on startup" } },
-                   { "s", { true,  "set session name" } },
-                   { "d", { false, "run as a headless session (requires -s)" } },
-                   { "p", { true,  "just send stdin as commands to the given session" } },
-                   { "f", { true,  "filter: for each file, select the entire buffer and execute the given keys" } },
-                   { "i", { true, "backup the files on which a filter is applied using the given suffix" } },
-                   { "q", { false, "in filter mode, be quiet about errors applying keys" } },
-                   { "ui", { true, "set the type of user interface to use (terminal, dummy, or json)" } },
-                   { "l", { false, "list existing sessions" } },
-                   { "clear", { false, "clear dead sessions" } },
-                   { "debug", { true, "initial debug option value" } },
-                   { "version", { false, "display kakoune version and exit" } },
-                   { "ro", { false, "readonly mode" } },
-                   { "help", { false, "display a help message and quit" } } }
+        SwitchMap{ { "c", { ArgCompleter{},  "connect to given session" } },
+                   { "e", { ArgCompleter{},  "execute argument on client initialisation" } },
+                   { "E", { ArgCompleter{},  "execute argument on server initialisation" } },
+                   { "n", { {}, "do not source kakrc files on startup" } },
+                   { "s", { ArgCompleter{},  "set session name" } },
+                   { "d", { {}, "run as a headless session (requires -s)" } },
+                   { "p", { ArgCompleter{},  "just send stdin as commands to the given session" } },
+                   { "f", { ArgCompleter{},  "filter: for each file, select the entire buffer and execute the given keys" } },
+                   { "i", { ArgCompleter{}, "backup the files on which a filter is applied using the given suffix" } },
+                   { "q", { {}, "in filter mode, be quiet about errors applying keys" } },
+                   { "ui", { ArgCompleter{}, "set the type of user interface to use (terminal, dummy, or json)" } },
+                   { "l", { {}, "list existing sessions" } },
+                   { "clear", { {}, "clear dead sessions" } },
+                   { "debug", { ArgCompleter{}, "initial debug option value" } },
+                   { "version", { {}, "display kakoune version and exit" } },
+                   { "ro", { {}, "readonly mode" } },
+                   { "help", { {}, "display a help message and quit" } } }
     };
 
     try
     {
-        auto show_usage = [&]()
-        {
+        auto show_usage = [&]() {
             write_stdout(format("Usage: {} [options] [file]... [+<line>[:<col>]|+:]\n\n"
                     "Options:\n"
                     "{}\n"
@@ -1230,7 +1250,7 @@ int main(int argc, char* argv[])
                              ((argc == 1 or (ignore_kakrc and argc == 2))
                               and isatty(0)                               ? ServerFlags::StartupInfo : ServerFlags::None);
                 auto debug_flags = option_from_string(Meta::Type<DebugFlags>{}, parser.get_switch("debug").value_or(""));
-                return run_server(session, server_init, client_init, init_coord, flags, ui_type, debug_flags, files);
+                return run_server(session, server_init, client_init, files.empty() ? StringView{} : files[0], init_coord, flags, ui_type, debug_flags, files);
             }
             catch (convert_to_client_mode& convert)
             {
